{% extends "base.html" %}

{% block announce %}
<a href="https://github.com/vstorm-co/pydantic-ai-middleware">
  <span class="twemoji">{% include ".icons/fontawesome/brands/github.svg" %}</span>
  <strong>Star us on GitHub</strong> — Help us reach more developers!
</a>
{% endblock %}
